var labelType, useGradients, nativeTextSupport, animate;

///////////////////////////////////////////////////
//	le main code
///////////////////////////////////////////////////

(function() {
  var ua = navigator.userAgent,
      iStuff = ua.match(/iPhone/i) || ua.match(/iPad/i),
      typeOfCanvas = typeof HTMLCanvasElement,
      nativeCanvasSupport = (typeOfCanvas == 'object' || typeOfCanvas == 'function'),
      textSupport = nativeCanvasSupport 
        && (typeof document.createElement('canvas').getContext('2d').fillText == 'function');
  //I'm setting this based on the fact that ExCanvas provides text support for IE
  //and that as of today iPhone/iPad current text support is lame
  labelType = (!nativeCanvasSupport || (textSupport && !iStuff))? 'Native' : 'HTML';
  nativeTextSupport = labelType == 'Native';
  useGradients = nativeCanvasSupport;
  animate = !(iStuff || !nativeCanvasSupport);
})();

var Log = {
  elem: false,
  write: function(text){
    if (!this.elem) 
      this.elem = document.getElementById('log');
    this.elem.innerHTML = text;
    this.elem.style.left = (500 - this.elem.offsetWidth / 2) + 'px';
  }
};

		
function Listener(){
}
Listener.prototype.onCall = function(node, eventInfo, e, listenerFunction){
	// do nothing
}

function GraphFlow(){
	
	/////////////////////////////////////////////
	//				VARIABLES				   //
	/////////////////////////////////////////////
	
	//var that = this;
	
	/** edge properties */
	var edgeColor 		= "#114270",
		edgeColorFocus 	= "#FF00FF";

	/** node properties */
	var vertexLabelSize 	= 12,		// this value modifies the text size and with it the node size
		nodeColor			= "#909090",
		nodeColorFocus		= "#FF00FF";
		//vertexLabelColor 	= "#0F0F0F",
		//vertexSelectColor	= "#B2E5F2";

	var longestEdge = 0;

	/** FlowGraph Object */
	var fd;

	/** Nodes and edges where the mouse moves over */
	var hover = null;

	/** This array stores all Nodes and is initialized with a dummy,
		used for beautiful edge dragging.
		*/
	var json = [ {
			  "adjacencies": [], 
			  "data": {
				"$dim": 0,
				"$type": "none",
				"&xPos": 0,
				"&typeSelected" : null,
				"&yPos": 0,
			  }, 
			  "id": "#DUMMY_MOUSE_NODE", 
			  "name": "",
			  "alpha": 0
		}];

	/** selected node:	{id : <the id of the selected node, 
						 from : <the data.$type of the selected node>}
						 
		When clicking a node or edge, this variable will contain node
		information, that is required for creating edges.
		*/
	var selectedNode = null;
	
	/** This points to the DOM-node (not the json-array node!) that is
		an invisible dummy for moving an edge with the mouse
		*/
	var mouseNode = null;

	/** true if the mouse is over a port-node/edge */
	var mouseOverNode = false;
	var mouseOverEdge = false;
	
	/** A hashtable of mouseEvents to arrays of functions that
		should happen on these events
		*/
	var listener = [];
	
	/////////////////////////////////////////////
	//				FUNCTIONS				   //
	/////////////////////////////////////////////

	/** 
		reloads the graph, trying to restore all positions 
		this function needs to be called whenever nodes or egdges are
		added or deleted!
	*/
	this.refresh = function(){
		// reload graph
		fd.loadJSON(json);
		// restore old node positions
		restoreNodePositions();
		fd.plot();
	}
	
	/**
		Adds a functionality to the following MouseEvents:
			onClick,
			onRightClick,
			onMouseMove,
			onMouseEnter,
			onMouseLeave,
			onDragStart,
			onDragMove,
			onDragCancel,
			onDragEnd
		When altering the graph, do not forget to refresh() it at the
		end!
		@param event - the name of the mouse event ( see list above)
		@param listenerFunction - a function that may alter the graph
	*/
	this.addListener = function(event, listenerFunction){
		if(listener[event] == undefined){
			listener[event] = [];
		}
		/*if(listenerFunction.arguments.length != 3){
			Log.write("Invalid Listener Function: needs 3 parameters");
			return;
		}*/
		// push new function to the listener table
		listener[event].push(listenerFunction);
	}
	
	/**
		Calls all listener that are registered under the eventName.
		@param eventName - see 'addListener()'
		@param node - the node which was clicked
		@param eventInfo - an object containing useful methods like *getPos* to get the mouse position relative to the canvas
		@param e - the grabbed event (should return the native event in a cross-browser manner)
	*/
	this.callListener = function(eventName, node, eventInfo, e){
		var lArr = listener[eventName];
		if(lArr == undefined){
			return;
		}
		// iterate through all listener this event concerns
		for(var l=0; l < lArr.length; l++){
			lArr[l](node, eventInfo, e);
		}
	}
	
	/**
	 Moves all nodes to their respective positions, which are stored within
	 their data. We need this function to dynamically remove graph elements
	 without losing the node positions.
	*/
	function restoreNodePositions(){
		for (var n = 0; n < json.length; n++){
			var x = json[n].data.$xPos,
				y = json[n].data.$yPos,
				node =  fd.graph.getNode(json[n].id);
			node.pos.setc(x,y);
		}
	}
	
	/**
	 Adds a Node to the graph.
	 @param familyID - the ID and displayname of the node
	 @param inputPortIDs - an array of IDs for the input ports
	 @param outputPortIDs - an array of IDs for the output ports
	 */
	this.addNode = function(xPosition, yPosition, familyID, inputPortIDs, outputPortIDs){
	
	var inputPorts = [],
		outputPorts = [],
		maxPorts = Math.max(inputPortIDs.length, outputPortIDs.length),
		size = vertexLabelSize*2;
		
	var width = (familyID.length+4)*vertexLabelSize,
		height = 1*size + 
			Math.max( 2*size, size/2 + maxPorts * size);
		
	// for fast position computation, calculate the relative position
	// of the first port to the nodeFamily-box
	var ipRelativeX = -width/2,
		ipRelativeY = (size - inputPortIDs.length * size)/2,
		opRelativeX = width/2,
		opRelativeY = (size - outputPortIDs.length * size)/2;
	
	// set position of the ports in world coordinates
	var x = xPosition+ ipRelativeX,
		y = yPosition+ ipRelativeY;
	
	// define inputPorts
	for(var p=0; p < inputPortIDs.length; p++){
		var inputPort =
			{
			  "adjacencies": [], 
			  "data": {
				"$dim": size,					
				"$type": "inputPort",
				"$xPos": x,
				"$yPos": y,
				"$color": "#AA0000",
				"$unfocusColor" : "#AA0000"
			}, 
			  "id": familyID+"_"+inputPortIDs[p], 
			  "name": inputPortIDs[p],
		};
		
		inputPorts.push(inputPort);
		y += size;
	}
	
	if(inputPorts.length != 0){
		inputPorts[0].data.$relativeX= ipRelativeX;
		inputPorts[0].data.$relativeY= ipRelativeY;
	}
	
	// set position of the ports in world coordinates
	x = xPosition+ opRelativeX,
	y = yPosition+ opRelativeY;
	
	// define outputPorts
	for(var p=0; p < outputPortIDs.length; p++){
		var outputPort =
			{
			  "adjacencies": [], 
			  "data": {
				"$dim": size,					
				"$type": "outputPort",
				"$xPos": x,
				"$yPos": y,
				"$color": "#0000AA",
				"$unfocusColor" : "#0000AA"
			}, 
			  "id": familyID+"_"+outputPortIDs[p], 
			  "name": outputPortIDs[p],
		};
		
		outputPorts.push(outputPort);
		y += size;
	}
	if(outputPorts.length != 0){
		outputPorts[0].data.$relativeX= opRelativeX;
		outputPorts[0].data.$relativeY= opRelativeY;
	}
	var nodeFamily =
		{
		  //"adjacencies": [], 
		  "data": {
			"$dim": size,					
			"$type": "nodeFamily",
			"$jsonIndex": json.length,
			"$portCount": inputPortIDs.length+ outputPortIDs.length,
			"$width": width,
			"$height": height,
			"$color": "#909090",
			"$unfocusColor" : "#909090",
			"$fillColor": "#F0F0F0",
			"$xPos": xPosition,
			"$yPos": yPosition
		  }, 
		  "id": familyID, 
		  "name": familyID,
		};
		
	// set position of the close button in world coordinates
	x = xPosition+ width/2,
	y = yPosition- height/2;
	
	var closeButton =
		{
		  "adjacencies": [], 
		  "data": {
			"$dim": size,					
			"$type": "crossBox",
			"$family": nodeFamily,
			"$xPos": x,
			"$yPos": y,
			"$color": "#CC0909",
			"$unfocusColor" : "#CC0909",
		   }, 
		  "id": familyID+"_close", 
		  "name": "x",
		};

	// add all components
	json.push(nodeFamily);
	json.push(closeButton);
	json= json.concat(inputPorts);
	json= json.concat(outputPorts);
	}

	/**
		Removes a Node and all its edges from the graph.
		@param nodeFamily - the node which is to be removed
	*/
	this.removeNode =  function(nodeFamily){
		
		var deleteFrom = nodeFamily.data.$jsonIndex, 
			deleteUntil = deleteFrom+ nodeFamily.data.$portCount+1;
		
		// delete nodeFamily and closeButton
		delete json[deleteFrom];
		delete json[deleteFrom+1];
		
		// delete ports
		for(var n = deleteFrom+2;  n <= deleteUntil;  n++){
			var portID = json[n].id;
			
			// remove all incoming edges
			for(var e = 0;  e < json.length;  e++){
				var incNode = json[e];
				if(incNode != undefined){
					removeEdge(incNode, portID);	
				}
			}
			
			delete json[n];
		}
		
		// collapse json array
		var deletionSum = deleteUntil-deleteFrom+1;
		json.splice(deleteFrom, deletionSum);
		selectedNode = null;
		
		// update nodeFamily indices
		for(var n = deleteUntil-deletionSum+1;  n < json.length;  n++){
			var familyIndex = json[n].data.$jsonIndex;
			if(familyIndex != undefined){
				json[n].data.$jsonIndex = n;
			}
		}
			
		return;
	 }
	 
	/**
		Moves a node to the specified position.
		@param nodeFamily - the node which is to be moved
		@param x - desired x-Position
		@param y - desired y-Position
	*/
	 function moveNode(nodeFamily, x, y){
		var moveFrom = nodeFamily.data.$jsonIndex, 
			moveUntil = moveFrom+ nodeFamily.data.$portCount+1,
			dim = nodeFamily.data.$dim;
			
		// set pos of family box
		nodeFamily.pos.setc(x, y);
		json[moveFrom].data.$xPos = x;
		json[moveFrom].data.$yPos = y;
		moveFrom++;
		
		// set pos of close button
		var node = fd.graph.getNode(json[moveFrom].id),
			nodeX = x+ nodeFamily.data.$width/2,
			nodeY =  y- nodeFamily.data.$height/2;
			
		node.pos.setc(nodeX, nodeY);
		json[moveFrom].data.$xPos = nodeX;
		json[moveFrom].data.$yPos = nodeY;
		moveFrom++;
			
		// set pos of inputs		
		nodeX = x+ json[moveFrom].data.$relativeX;
		nodeY = y+ json[moveFrom].data.$relativeY;
		
		for(var i= moveFrom; i<= moveUntil; i++){
			node = fd.graph.getNode(json[i].id);
			if(node.data.$type == "inputPort"){
				node.pos.setc(nodeX,nodeY);
				json[i].data.$xPos = nodeX;
				json[i].data.$yPos = nodeY;
				nodeY+= dim;
			}
			else{
				moveFrom = i;
				break;
			}
		}
		
		// set pos of outputs
		nodeX = x+ json[moveFrom].data.$relativeX;
		nodeY = y+ json[moveFrom].data.$relativeY;
		
		for(var o= moveFrom; o <= moveUntil; o++){
			node = fd.graph.getNode(json[o].id);
			node.pos.setc(nodeX,nodeY);
			json[o].data.$xPos = nodeX;
			json[o].data.$yPos = nodeY;
			
			nodeY+= dim;
		}
	 }

  
  /**
   *  adds an edge to the graph
   *  @param sourceID the id of the node from which the edge starts
   *  @param targetID the id of the node where the edge ends
   *  @return true if the edge was successfully added
   */
	this.addEdge = function(sourceID, targetID){
		// edge from node to itself?
		if(sourceID == targetID){
			Log.write("Error: Cannot connect Node to itself!");
			return false;
		}
		
		// look up the source node
		var source = getNodeFromJSON(sourceID),
			adja = source.adjacencies;
		
		// source node does not exist?
		if(adja == undefined){
			Log.write("Error: "+sourceID+" does not exist!");
			return false;
		}
	
		for(var a=0; a < adja.length; a++){
			// does the edge exist already?
			if(adja[a].nodeTo == targetID){
				Log.write("Error: Edge("+ adja[a].nodeFrom+", "+adja[a].nodeTo+") exists already!" );
				return false;
			}
		}
		var edge ={
				  "nodeTo": targetID, 
				  "nodeFrom": sourceID,
				  "data": {"$direction" : [ sourceID, targetID ] }
				  };
				  
		if(sourceID == mouseNode.id || targetID == mouseNode.id){
			edge.data.$type = "mouseArrow";
		}
		
		adja.push(edge);
		return true;
	}
  
  /**
	Removes the edge between two vertices.
	@param source the source Node
	@param targetID the id of the target Node
	*/
  this.removeEdge = function(source, targetID){
	var adja = source.adjacencies;
	
	// nothing to delete?
	if(adja == undefined){
		return;
	}
	for(var a=0; a< adja.length; a++){
		if(adja[a].nodeTo == targetID){
			delete adja[a];
			adja.splice(a,1);
			break;
		}
	}
  }
  
  /**
	makes it so that a text label is not selectable.
	(unfortunately, this does not prevent CTRL+A selection)
	*/
  function setLabelUnselectable(label){
	// internet explorer:
	if (typeof label.onselectstart!="undefined"){
        label.onselectstart =
			function(){return false};
		}
	// mozilla firefox:
    else if (typeof label.style.MozUserSelect!="undefined"){
        label.style.MozUserSelect = "none";
	}
	// others:
    else{
        label.onmousedown =
			function(){return false};
	}
  }
  
    /**
	saves the given x, y position of a node
	@param nodeID - the id of the target node
	@param x - x-position of the node
	@param y - y-position of the node
	*/
  function saveNodePosition(nodeID, x, y){
	var node =getNodeFromJSON(nodeID);
	node.data.$xPos = x;
	node.data.$yPos = y;
  }
  
  /**
  returns a jsonNode by id
  @param nodeID the id of the desired node
  */
  function getNodeFromJSON(nodeID){
	for(var n=0; n< json.length; n++){
		if(json[n].id == nodeID){
			return json[n];
		}
	}
	return null;
  }
  
  /**
	clears all highlights, caused by mouse-enter events
	@param clearNodes - if true, highlighted nodes will be cleared
	@param clearEdges - if true, highlighted edges will be cleared
	*/
  function setHighlight(node){
		// animation parameters
		var trans = $jit.Trans.Quint.easeOut, 
			dur= 200;
			
		// clean up the old highlight
		if(hover != null){
			
			if(hover.nodeFrom != undefined){
				hover.setData('lineWidth', 1, 'end');
				hover.setData('color', edgeColor, 'end');
			}
			else{
				hover.setData('color', hover.data.$unfocusColor, 'end');
			}
			hover = null;
		}
		
		// do we highlight something new?
		if(node != null){
			if(node.nodeFrom != undefined){
				node.setData('color', edgeColorFocus, 'end');
				node.setData('lineWidth', 2, 'end');
			}else{
				node.setData('color', nodeColorFocus, 'end');
			}
			hover = node;
		}else{
			// slower fade away transition
			trans = $jit.Trans.Quint.easeIn;
			dur = 400;
		}
		
		// Animations overwrite each other. If we let this animation
		// execute while we delete an edge, the onComplete() function
		// of the edge animation will not be called
		fd.fx.animate({  
				modes: ['edge-property:lineWidth:color',
						'node-property:color'],
				transition: trans,  
				duration: dur
		}); 
  }
    
  
  /**
	Initializes the Graph
	*/
  this.initGraph = function(graph){
  
	  if(graph != null){
		json = graph;
	  }
	  
	  fd = new $jit.FlowGraph({
		//id of the visualization container
		injectInto: 'infovis',
		//Enable zooming and panning
		//with scrolling and DnD
		Navigation: {
		  enable: true,
		  //Enable panning events only if we're dragging the empty
		  //canvas (and not a node).
		  panning: 'avoid nodes',
		  zooming: 10 //zoom speed. higher is more sensible
		},
		// Change node and edge styles such as
		// color and width.
		// These properties are also set per node
		// with dollar prefixed data-properties in the
		// JSON structure.
		Node: {
		  overridable: true,
		  color: nodeColor,
		  CanvasStyles: {  
			  lineWidth: 2  
			}  
		  
		},
		Edge: {
		  overridable: true,
		  color: edgeColor,
		  lineWidth: 1,
		  dim: vertexLabelSize,
		  type: 'flowArrow'
		},
		//Native canvas text styling  
		Label: {  
			type: "HTML", //Native or HTML  
			size: 10,  
			style: 'bold'  
		},  
		// Add node events
		Events: {
		  enable: true,
		  enableForEdges: true,
		  type: 'Native',
		  
		  /**
		   *	EVENT:	OnRightClick
		   */
		  onRightClick: function(node, eventInfo, e) {
			callListener("onRightClick", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnDragEnd
		   */
		  onDragEnd: function(node, eventInfo, e) {
			callListener("onDragEnd", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnDragCancel
		   */
		  onDragCancel: function(node, eventInfo, e) {
			callListener("onDragCancel", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnMouseEnter
		   */
		  onMouseEnter: function(node, eventInfo, e) {
			if(node.id == mouseNode.id){
				return;
			}
			
			// ignore some highlights if we are dragging an edge
			if(selectedNode != null && 
				(node.data.$type != "inputPort" && node.data.$type != "outputPort" )){
				return;
			}
			
			// check if object is an edge or a node
			mouseOverNode = (node.nodeFrom == undefined);
			mouseOverEdge = !mouseOverNode;
			
			setHighlight(node);
			
			if(node.data.$type == "nodeFamily"){
				fd.canvas.getElement().style.cursor = 'move';
			}else{
				fd.canvas.getElement().style.cursor = 'pointer';
			}
			callListener("onMouseEnter", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnMouseMove
		   */
		  onMouseMove: function(node, eventInfo, e) {
			// are we dragging an edge with the mouse ?
			if(selectedNode != null){
				var pos = eventInfo.getPos();
				mouseNode.pos.setc(pos.x, pos.y-2);
				fd.plot();
			}
			callListener("onMouseMove", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnMouseLeave
		   */
		  onMouseLeave: function(node, eventInfo, e) {
			fd.canvas.getElement().style.cursor = '';
			mouseOverNode = false;
			mouseOverEdge = false;
			
			setHighlight(null);
			callListener("onMouseLeave", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnDragStart
		   */
		  onDragStart: function(node, eventInfo, e) {
			if(node == false || node.id == mouseNode.id ||
			   node.data.$type != "nodeFamily"){
				return;
			}
			
			// memorize where we drag the node instead on centering
			// it to the mouse pointer
			var mousePos = eventInfo.getPos(),
				nodePos = node.pos.getc(true);
				
			mouseNode.data.$dragX = nodePos.x - mousePos.x;
			mouseNode.data.$dragY = nodePos.y - mousePos.y;
			callListener("onDragStart", node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnDragMove
		   */
		  onDragMove: function(node, eventInfo, e) {
			if(node == false || node.id == mouseNode.id ||
			   node.data.$type != "nodeFamily"){
				return;
			}
			var pos = eventInfo.getPos();
			
			moveNode(node, 
						   pos.x+mouseNode.data.$dragX,
							pos.y+mouseNode.data.$dragY);
			callListener("onDragMove", node, eventInfo, e);
			fd.plot();
			
		  },
		  
		  /**
		   *	EVENT:	OnTouchMove
		   */
		  onTouchMove: function(node, eventInfo, e) {
			$jit.util.event.stop(e); //stop default touchmove event
			this.onDragMove(node, eventInfo, e);
		  },
		  
		  /**
		   *	EVENT:	OnClick
		   */
		  onClick: function(node, eventInfo, e){
		  var mousePos = eventInfo.getPos();
		  
		  if(mouseOverNode && node != false){
		  
			switch(node.data.$type){
				case "crossBox":
					removeNode(node.data.$family);
					refresh();
					break;
				case "nodeFamily":
					break;
				case "outputPort":
					// no selection yet
					if(selectedNode == null){
						selectedNode = {"id": node.id, "from": "outputPort"};
						
						// add edge to mouseNode
						addEdge(node.id, mouseNode.id);
						refresh();
						return;
					}
					// click outside the edge drag box
					else{
						// add Edge if the selectedNode differs from the clickedNode
						if(selectedNode.from == "outputPort"){
							Log.write("Error: Cannot connect two ports of the same type!");
							break;
						}
						var newEdgeAdded = addEdge(node.id, selectedNode.id);
						if(newEdgeAdded){
							refresh();
						}
					}
					break;
				case "inputPort":
					// no selection yet
					if(selectedNode == null){
						selectedNode = {"id": node.id, "from": "inputPort"};
						
						// add edge to mouseNode
						addEdge(mouseNode.id, node.id);
						refresh();
						return;
					}
					// click outside the edge drag box
					else{
						// add Edge if the selectedNode differs from the clickedNode
						if(selectedNode.from == "inputPort"){
							Log.write("Error: Cannot connect two ports of the same type!");
							break;
						}
						var newEdgeAdded = addEdge(selectedNode.id,node.id);
						if(newEdgeAdded){
							refresh();
						}
					}
					break;
			}
			
		  }
		  else if(mouseOverEdge){
			// no selection yet
				if(selectedNode == null){
					selectedNode = {"id":node.data.$direction[0], "from":"outputPort"};//nodeTo
					
					var selectedNodeJson = getNodeFromJSON(selectedNode.id);
					
					Log.write(selectedNode.id+"->"+node.data.$direction[1]);
					// remove selectedEdge
					removeEdge(selectedNodeJson, node.data.$direction[1]);
					
					// add edge to mouseNode
					addEdge(selectedNode.id, mouseNode.id);
					refresh();
					return;
				}
		  }
		  
		  // remove selection and mouse-edge upon clicking anywhere
		  if(selectedNode != null){
		  
			if(selectedNode.from == "outputPort"){
				var selectedNodeJson = getNodeFromJSON(selectedNode.id);
				removeEdge(selectedNodeJson, mouseNode.id);
			}else{
				removeEdge(json[0], selectedNode.id);
			}
			
			selectedNode = null;
			refresh();
		  }
		  callListener("onClick", node, eventInfo, e);
		  }
		},
		//Number of iterations for the FD algorithm
		iterations: 200,
		//Edge length
		levelDistance: (longestEdge+2)*vertexLabelSize,
		// This method is only triggered
		// on label creation and only for DOM labels (not native canvas ones).
		onCreateLabel: function(domElement, node){
			// do not add labels if it is a dummy element
			// instead save pointer to domElement
			if(node.id == "#DUMMY_MOUSE_NODE"){
				mouseNode = node;
				return;
			}
			if(node.data.$type != "nodeFamily"){
				return;
			}
			
			// save node position
			saveNodePosition(node.id, node.pos.x, node.pos.y);
			
			domElement.innerHTML = node.name;  
			var style = domElement.style;  
			/*var nameContainer = document.createElement('span'),
				  styleName = nameContainer.style;
				  
			  nameContainer.className = 'fTitle';
			  nameContainer.innerHTML = node.name;
			  setLabelUnselectable(nameContainer);
			  domElement.appendChild(nameContainer);*/
			// style for nameContainer
			style.fontSize = 1.65*vertexLabelSize+"px";
			style.left = 0.2*vertexLabelSize+"px";
			style.color = node.data.$color;
		 
		},
		
		// Change node styles when DOM labels are placed
		// or moved.
		onPlaceLabel: function(domElement, node){
		  var style = domElement.style;
		  var left = parseInt(style.left);
		  var top = parseInt(style.top);
		  var w = domElement.offsetWidth;
		  var h = domElement.offsetHeight;
		  style.left = (left - w / 2) + 'px';
		  style.top = (top - h / 2) + 'px';
		  style.display = '';
		}
	  });
	  
	  // load JSON data.
	  fd.loadJSON(json,0);
	  // compute positions
	  refresh();
	  // end
	  //alert("final wut:"+fd.nodeSizeModifier);
	}
	
	return this;
}

// test data
function init(){
	var addNodeCounter = 0;
	var graph = GraphFlow();
	graph.addListener("onRightClick", function(node,info,e){
										graph.addNode(info.getPos().x, info.getPos().y, "lol"+addNodeCounter, ["ip1"], ["op1"]); 
										graph.refresh();
										addNodeCounter++;
					  });
	
	graph.addNode(0, -100, "superNode1", ["ip1","ip2","ip3","ip4","ip5","ip6"], ["op1"]);
	graph.addNode(0, 100, "superNode2", ["ip1"], ["op1","op2"]);
	graph.initGraph(null);
	graph.addNode(100, 0, "superNode3", ["ip1"], ["op1","op2"]);
	graph.refresh();
}
